from .spatial import *
from .storage import *
from .users import *
from .spatial.match import SpatialMatcher, get_nuts_matcher
from .storage.table import TablePersister
from .storage.file import AzureShareClient
from .storage.blob import AzureBlobClient

__all__ = ['spatial', 'storage', 'users', 'formats', 'io', 
           'models', 'utils', 'SpatialMatcher', 'get_nuts_matcher',
           'TablePersister', 'AzureShareClient', 'AzureBlobClient']


from .aad import  (get_application_users, get_users, get_user_application_roles, get_user_anagraphics,
                  get_token_auth_header, parse_client_principal,
                  read_user_authorization, read_user_from_jwt, 
                  msgraph_auth, msgraph_request)



from .match import SpatialMatcher, get_nuts_matcher, build_nuts_matcher
from .utils import get_earth_radius, convert_goem, distance_to_bounds

__all__ = ['match', 'utils', 'get_nuts_matcher', 'build_nuts_matcher', 'get_earth_radius', 'convert_goem', 'distance_to_bounds']


import base64
import json
import math
import numpy as np
from datetime import date, timedelta, datetime, timezone
import pandas as pd
import pytz

from creaap.formats import to_tz_aware_datetime

class EntityWithViewVariables():
	'''class of objects that have interla variables calculated
	after their initialization'''
	def build_view_variables(self):
		pass


# handling JSON serialization fo datetime objects
def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if type(obj) is float and math.isnan(obj):
        return "NaN"
    if isinstance(obj, np.int64): 
        try:
            return int(obj)
        except:
            return "NaN"
    elif isinstance(obj, (datetime, date)):
        return obj.isoformat()
    elif isinstance(obj, EntityWithViewVariables):
        obj.build_view_variables()
        return vars(obj)
    else:
        try:
            return obj.__dict__
        except:
            raise TypeError ("Type %s not serializable" % type(obj)) 
		
def __get_param_from_az_req_form(req, par_name):
	'''As Azure requests also have form data, this reads from said form data'''
	return req.form.get(par_name)

def get_parameter_with_default(req, par_name, default_val = None):
	'''
	Parameters
	----------
	req:
		an http request from an Azure Function
	par_name:str
		how the parameter is supposed to be named in the http request
	default_val: whatever
		a fallback value, defaults to None
	
	Returns
	-------
	object
		whatever the httprequest contained under par_name
		or whatever fuckery you passed as default value
	'''
	try:
		value = get_parameter(req, par_name)
	except ValueError:
		value = default_val
	return value

def get_parameter(req, par_name):
	'''
	Parameters
	----------
	req:
		an http request from an Azure Function
	par_name:str
		how the parameter is supposed to be named in the http request

	Returns
	-------
	object
		whatever the httprequest contained under par_name
	'''
	value = req.params.get(par_name)
	if value is None:
		value = __get_param_from_az_req_form(req, par_name)
		if value is None:
			try:
				req_body = req.get_json()
			except ValueError:
				raise ValueError(str(par_name) + ' cannot be found in request')
			else:
				value = req_body.get(par_name)
				if value is None:
					raise ValueError(str(par_name) + ' cannot be None')
	return value

def get_datetime_parameter_with_default(req, par_name:str, default_val:datetime, timezone_par:str = None, target_timezone = pytz.utc, **kwargs)->datetime:
	'''
	Parameters
	----------
	req:
		a http request from an Azure Function
	par_name:str
		how the datatime parameter is supposed to be named in the http request
	default_val:datetime
		a fallback datetime value
	timezone_par:str
		the name the timezone paramter has in the http request, if any.
		If this paramter is not provided and the datetime in the htto request is naive,
		the function will assume it is in the same timezone as the target_timezone.
	target_timezone:str ot pytz timezone
		the desired output timezone

	Returns
	-------
	datetime
		a timezone-aware datetime object
		or whatever shit you entered as default_val

	'''
	val = get_parameter_with_default(req, par_name, default_val = None)
	# parse the timezone parameter, if any
	if timezone_par:
		tz = get_parameter_with_default(req, timezone_par, default_val = target_timezone)
	else:
		tz = target_timezone
	if val:
		try:
			return to_tz_aware_datetime(val, timezone = tz, fallback=default_val)
		except:
			return default_val
	return default_val





'''Classes and functions to upload and download stuff from Azure File shares'''
import io
from azure.storage.fileshare import ShareServiceClient, ShareFileClient
from datetime import datetime
import pytz

from creaap.formats import denaive_datetime, to_tz_aware_datetime

class AzureShareClient:
    '''A simple client to perform upload and download operations from file share'''
    def __init__(self, connection_string,share_name):
        '''
        Creates a connection client to a specific Azure File Share

        Parameters
        ----------
        conenction_string: string
            the Azure Storage connection string. Grab it in the Azure Portal
        
        share_name: string
            the name of the File Share you want to connect to
        '''
        share_client = ShareServiceClient.from_connection_string(connection_string)
        self.client = share_client.get_share_client(share_name)

    def ls_files(self, path, namestart = None, recursive = False):
        '''
        List files under a path, optionally recursively.

        Paramters
        ---------
        path: string
            where to start file system exploration

        namestart: string, default None
            a prefix to filter files.

        recursive: bool, default False
            whether or not you want your ls to be recursive, i.e.
            traverse all subdirectories as well.

        Returns
        -------
        generator
            strings that represent File Share paths

        '''
        if not namestart == '':
            namestart == namestart
        share_path = self.client.get_directory_client(path)
        share_iter = share_path.list_directories_and_files(name_starts_with=namestart)
        for file in share_iter:
            yield file.name
            # in case we want a recursive ls
            if recursive and file.get('is_directory'):
                yield from self._ls_subdir(share_path.get_subdirectory_client(file['name']), namestart)
            

    def _ls_subdir(self, subdir_client, namestart = None):
        '''helper method to implement recursive LS'''
        share_iter = subdir_client.list_directories_and_files(name_starts_with=namestart)
        for file in share_iter:
            yield file.name
            # in case we want a recursive ls
            if file.get('is_directory'):
                yield from self._ls_subdir(subdir_client.get_subdirectory_client(file['name']), namestart)


    def get_updates(self, base_path = '', start_date = datetime.utcnow()):
        '''returns all the files that have been updated/added down the specified base_path after the start_time.
        Useful to poll file shares as currently Azure provides no "file share trigger"
        
        Parameters
        ----------
        base_path: string, default ''
            the directory where to look for updates. By default it's the root directory, so
            it will get any update in the whole file share.

        start_date: datetime.datetime, default right now in UTC time
            the first datetime that qualifies a file as "updated". It should be the
            datetime you last checked for updates.

        Returns
        -------
        generator
            a generator of azure.storage.fileshare.ShareFileClient objects
            to handle the updated files.
        
        '''
        # a little work on inputs...
        try:
            root_dir_client = self.client.get_directory_client(base_path)
        except:
            raise ValueError('Specified base path does not exist')
        # check if we have a legit datetime object
        if isinstance(start_date, datetime):
            # just to be sure it has a timezone attached
            start_date = denaive_datetime(start_date)
        else:
            start_date = to_tz_aware_datetime(start_date)
        # now on to fetching 
        yield from self._recursive_dir_updates(root_dir_client, start_date)
                
    def _recursive_dir_updates(self, dir_client, start_date):
        '''support method to recursively traverse the remote file system and get updated files'''
        for x in dir_client.list_directories_and_files():
            if x.get('is_directory'):
                yield from self._recursive_dir_updates(dir_client.get_subdirectory_client(x['name']), start_date)
            else:
                file_client = dir_client.get_file_client(file_name=x['name'])
                # get file's last upudate. Azure file sharesa work in UTC.
                load_time = denaive_datetime(file_client.get_file_properties()['creation_time'], timezone = pytz.utc)
                # tz-aware datetime can be compared even if in different timezones, so there is no need to
                # convert everything to UTC
                if (load_time > start_date):
                    yield file_client

    def download_to_stream(self, file):
        '''Downloads a File Share file into a BytesIO object
        
        Parameters
        ----------
        source: string or azure.storage.fileshare.ShareFileClient
            file path of the target resource, or ShareFileClient object connected to it

        Returns
        -------
        BytesIO
            byte-object representation of the blob resource
        '''
        
        if isinstance(file, ShareFileClient):
            file_client = file
        elif isinstance(file, str):
            file_client = self.client.get_file_client(file_path = file)
        else:
            raise ValueError('the *file* parameter has to be either a string representing a path to a file in the share or a ShareFileClient object')
        data = file_client.download_file()
        stream = io.BytesIO()
        data.download_to_stream(stream, max_concurrency=1)
        # go back to the beginning of the data stream
        stream.seek(0)
        return stream

    def upload_stream(self, data, dest):
        '''uploads a bytes object to the specified File Share path
        
        Parameters
        ----------
        data: bytes
            the object you want to upload

        dest: string
            te file share path where you want to slap the file
        '''
        file_client = self.client.get_file_client(file_path = dest)
        file_client.upload_file(data)

    def upload(self, source, dest):
        '''uploads a local file to the specified File Share path
        
        Parameters
        ----------
        source: string
            the local filepath

        dest: string
            te file share path where you want to slap the file
        '''
        with open(source, "rb") as source_file:
            self.upload_stream(source_file, dest)


from datetime import datetime
import json
import inspect
from creaap.formats import AZURE_TABLES_DATETIME_FORMAT, datetime_to_tables_string, to_table_storage_key, unfuck_numpy_formats
from creaap.io import EntityWithViewVariables, json_serial


class DomainEntity(EntityWithViewVariables):
    '''Parent class for all domain entities, allows rapid serialization of objects
    
     Class Attributes
    ----------------
    table: string
        the name of the correnspondig table in the Azure Tables persistence scheme.
        If set to None, the entity is not persisted to Table Storage

    row_key: string
        the name of the entity's attribute that will be used as RowKey

    partition_key: string
        the name of the entity's attribute that will be used as PartitionKey

    date_index_format: string
        in case your entity has a datetime as either Row or Partition Key, here
        you can specify how it should look like in the table storage, becasue
        keys must be strings. By default it's "%Y-%m-%dT%H:%M:%S.%fZ"
        Mind also that your datetimes will be converted to UTC to avoid 
        duplication, which may cause clashes.
    '''
    table = 'Entity'
    partition_key = 'PartitionKey'
    row_key = 'RowKey'
    date_index_format = AZURE_TABLES_DATETIME_FORMAT

    @classmethod
    def from_dict(cls, env):      
        return cls(**{
            k: v for k, v in env.items() 
            if k in inspect.signature(cls).parameters
        })

    def __eq__(self, other):
        '''Two domain entities are to be considered equal if they clash once put in the table storage'''
        if isinstance(other, DomainEntity):
            if other.table == self.table:
                    return (self.__dict__[self.row_key] == other.__dict__[other.row_key]) & (self.__dict__[self.partition_key] == other.__dict__[other.partition_key])
        return False
    
    def __hash__(self):
        '''Since we cusotmized the __eq__ method, we need to declare a new __hash__ one too
        to allor for hashable domain entities'''
        return hash(self.table + self.partition_key + self.row_key)
    
    def to_json(self):
        '''returns a JSON representaion of the current object'''
        return json.dumps(self, default=json_serial, 
            sort_keys=True, indent=4)

    def build_view_variables(self):
        '''By default it does nothing, but it is mant to build additional
        custom variables that are not really serialized, like active/inactive
        flags for entities that have expiration dates. It gets called when
        a DomainEntity is serialized into a JSON object
        '''
        pass

    def to_table_entities(self):
        '''Return a generator containing DB-manageable objects that represent the current domain entity'''
        out= {}
        for key, value in self.__dict__.items():
            if (key == self.partition_key):
                out['PartitionKey'] = to_table_storage_key(value, date_format = self.date_index_format)
            elif key == self.row_key:
                out['RowKey'] = to_table_storage_key(value, date_format = self.date_index_format)
            elif isinstance(value, DomainEntity):
                yield from value.to_table_entities()
            elif isinstance(value, list) |  isinstance(value, tuple):
                for elem in value:
                    try:
                        yield from elem.to_table_entities()
                    except:
                        print(' Object ' + str(elem) + ' in field ' + str(key) + ' is not a valid domain entity')
            else:
                # here we are dealing with "anything else". Which can be some Numpy fuckery
                # that Table Storage won't be able to digest. Luckily we have a failsafe 
                # method to prevent that.
                out[key]= unfuck_numpy_formats(value)
        yield self.table, out
    
    def expand(self):
        '''fetches all possible connected information'''
        pass


class DataBatch(DomainEntity):
    '''
    Commodity entity to move around batches of Domain Entities.

    DomainEntities in the DataBatch can be of different types and the 
    Databatch can be flushed to the Table Storage like any other 
    DomainEntity object and all its components will be inserted/update/
    upserted.

    By default, the 'table' class attribute is set to None to prevent
    the DataBatch itself from being persisted, however, if you need to,
    you can subclass it into persistable entities.
    '''
    table = None
    def __init__(self):
        self.batch = []

    def append(self,data):
        '''Adds a data object to the container'''
        if isinstance(data, DomainEntity):
            self.batch.append(data)
        else:
            raise TypeError("Only valid DomainEntity types can be added to a DataBatch")
    
    def to_table_entities(self):
        '''Translates the container into table entites'''
        for x in self.batch:
            yield from x.to_table_entities()
	
    def __iter__(self):
        for i in self.batch:
            yield i

    def __len__(self):
        return len(self.batch)

class SpatialEntity(DomainEntity):
    '''
    Class of geo-referenced entities
    This class should be regareded as an abstract one and used as base for defining 
    your own entities that come with an associated geographical information
    
    Class Attributes
    ----------------
    table: string
        the name of the correnspondig table in the Azure Tables persistence scheme.
        If set to None, the entity is not persisted to Table Storage

    row_key: string
        the name of the entity's attribute that will be used as RowKey

    partition_key: string
        the name of the entity's attribute that will be used as PartitionKey

    geometry: string
        the name of the entity's attribute that will be used to build the spatial index
    '''
    table = None
    partition_key = 'PartitionKey'
    row_key = 'RowKey'
    geometry_attribute = 'geometry'

class BlobEntity(DomainEntity):
    '''
    Class for entities with attached blob objects.
    This class should be regareded as an abstract one and used as base for defining 
    your own entities that come with an associated Blob file
    
    Class Attributes
    ----------------
    table: string
        the name of the correnspondig table in the Azure Tables persistence scheme

    row_key: string
        the name of the entity's attribute that will be used as RowKey
        If set to None, the entity is not persisted to Table Storage

    partition_key: string
        the name of the entity's attribute that will be used as PartitionKey

    blob_path_attribute: string
        the name of the entity's attribute that will be used as PartitionKey

    '''
    table = None
    partition_key = 'PartitionKey'
    row_key = 'RowKey'
    blob_path_attribute = 'blob'




from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
   name='creaap',
   version='0.1.15',
   description='CREA Applications in Python toolkit. A bundle of tools to speed up the deelopment of new Python applications in Azure. Allows you to abstract a little over persistence and to manage spatial data.',
   author='CREA',
   author_email='dario.denart@crea.gov.it',
   packages=['creaap', 'creaap.spatial', 'creaap.storage', 'creaap.users'],
   classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
   install_requires=['azure-storage-blob','azure-storage-file-share','azure-data-tables','Shapely>=2.0.0','pytz','pygeoif',
                     'numpy','scipy','scikit-learn','numpy','pyshp','pandas','pyproj', 'msal'], #external packages as dependencies
   include_package_data=True, # to be able to access data files
   long_description=long_description,
   long_description_content_type="text/markdown",
   url="https://dev.azure.com/dariodenart/_git/CREA%20Application%20Python%20toolkit"
)


'''Classes and functions to store and query data from Azure Tables'''
# Reference: https://docs.microsoft.com/en-us/python/api/overview/azure/data-tables-readme?view=azure-python
import logging
from azure.data.tables import TableServiceClient
from azure.core.exceptions import ResourceExistsError
from creaap.models import DomainEntity

from creaap.utils import get_all_subclasses, id_generator, first


class TablePersister():
	'''
	Allows to handle data on Table Storage like it was on a
	object-based DB. Sort of. BAsically allows you to use 
	Table Storage as a DB replacement.
	'''


	def __init__(self, connection_string):
		'''Intantiates a Table Storage connection client
		
		Parameters
        ----------
        conenction_string: string
            the Azure Storage connection string. Grab it in the Azure Portal
		'''
		self.table_service_client = TableServiceClient.from_connection_string(conn_str=connection_string)
		# this is a list of classes in the Models module
		clsmembers = get_all_subclasses(DomainEntity)
		# and now we turn it into a dictionary of DB table -> DomainEntity associations
		self.class_index={}
		for model_class in clsmembers:
			if issubclass(model_class, DomainEntity):
				if model_class.table:
					self.class_index[model_class.table] = model_class

	def _to_domain_entity(self, dictionary, table):
		'''Helper method to turn an Azure Tables row into a Domain Entity'''
		domain_class = self.class_index[table]
		try:
			dictionary[domain_class.partition_key] =  dictionary.pop("PartitionKey")
		except:
			dictionary[domain_class.partition_key] = None
		try:
			dictionary[domain_class.row_key] =  dictionary.pop('RowKey')
		except:
			dictionary[domain_class.row_key] = None
		return domain_class.from_dict(dictionary)

	def init_tables(self):
		'''
		Considers all defined DomainEntities in our project and creates
		a brand new table for each one of them.

		Returns
		-------
		None
		'''
		for table, _ in self.class_index.items():
			try:
				self.create_table(table)
			except ResourceExistsError:
				logging.warning("Nice try, but table " + table + ' already exists, asshole!')


	def create_table(self, table_name):
		'''creates a new table.
		
		Parameters
		----------
		table_name: string
			the name of the new table. Don't worry if it already esists

		Returns
		-------
		None
		'''
		return self.table_service_client.create_table(table_name=table_name)

	def insert_entity(self, domain_entity):
		'''inserts a new entity and all its components
		
		Parameters
		----------
		domain_entity: creaap.models.DomainEntity
			the object you need to be persisted

		Returns
		-------
		generator
			a generator of azure.data.tables.TableItem objects, returning a TableItem
			for each table row affected by the operation.
		'''
		for table, entity in domain_entity.to_table_entities():
			with self.table_service_client.get_table_client(table_name=table) as table_client:
				yield table_client.create_entity(entity=entity)

	def update_entity(self, domain_entity):
		'''updates a new entity and all its components
		
		Parameters
		----------
		domain_entity: creaap.models.DomainEntity
			the object you need to be persisted
		
		Returns
		-------
		generator
			a generator of azure.data.tables.TableItem objects, returning a TableItem
			for each table row affected by the operation.
		'''
		for table, entity in domain_entity.to_table_entities():
			with self.table_service_client.get_table_client(table_name=table) as table_client:
				yield table_client.update_entity(entity=entity)

	def upsert_entity(self, domain_entity):
		'''upserts a new entity and all its components
		
		Parameters
		----------
		domain_entity: creaap.models.DomainEntity
			the object you need to be persisted
		
		Returns
		-------
		generator
			a generator of azure.data.tables.TableItem objects, returning a TableItem
			for each table row affected by the operation.
		'''
		for table, entity in domain_entity.to_table_entities():
			with self.table_service_client.get_table_client(table_name=table) as table_client:
				yield table_client.upsert_entity(entity=entity)

	def query_entity(self, table, query, parameters = {}, **kwargs):
		'''Executes Azure Table queries
		
		Parameters
		----------
		table: string
			the table you intend to query.

		query: string
			the Azure Tables query string.

		paramentries: dict, default {}
			optional parameters dictionary for the above query
		
		Returns
		-------
		generator
			a generator of creaap.models.DomainEntity, returning a DomainEntity
			for each table row retireved by the operation.
		'''
		with self.table_service_client.get_table_client(table) as tc:
			for x in tc.query_entities(
						query_filter=query, parameters=parameters, **kwargs
					):
				yield self._to_domain_entity(x, table)
	
	def get_all_entities(self, table):
		'''Retrieves all records from a given table. 
		It's like a SELECT * FROM table

		Parameters
		----------
		table: string
			the table you intend to retrive in its interity.
		
		Returns
		-------
		generator
			a generator of creaap.models.DomainEntity, returning a DomainEntity
			for each table row retireved by the operation.
		'''
		with self.table_service_client.get_table_client(table_name=table) as table_client:
			for _, row in enumerate(table_client.list_entities()):
				yield self._to_domain_entity(row, table)

	def delete_entity(self, table, partition_key, row_key):
		'''Permanently deletes an entiy from a table
		
		Parameters
		----------
		partition_key: string
			the PartitionKey of the object you want to delete

		row_key: string
			the RowKey of the object you want to delete
		
		Returns
		-------
		generator
			a generator of azure.data.tables.TableItem objects, returning a TableItem
			for each table row affected by the operation.
		'''
		with self.table_service_client.get_table_client(table_name=table) as table_client:
			yield table_client.delete_entity(partition_key, row_key)

	def commit_large(self, domain_entity, operation = "insert", options ={}):
		'''This commits bulk operations on big entities with a truckload of smaller entities inside.
		
		Parameters
		----------
		domain_entity: creaap.models.DomainEntity
			the object you need to be persisted

		operation: string
			the operation you'd like to perform, one of the following: 'insert', 'upsert',
			'update', 'delete'
		
		options: dict, default {}
			the optional parameter dictionary to adderss thing such as the merge mode 
			for update/upsert operations.
		
		Returns
		-------
		generator
			a generator of azure.data.tables.TableItem objects, returning a TableItem
			for each table row affected by the operation.
		'''
		# BIG problem: batches MUST have the same Partition Key
		to_commit = {}
		for table, entity in domain_entity.to_table_entities():
			if to_commit.get(table):
				# update the existing dictionary
				to_commit[table]= self._separate_pks(entity['PartitionKey'], (operation, entity, options), to_commit[table])
			else:
				# create a whole new one
				to_commit[table]= self._separate_pks(entity['PartitionKey'], (operation, entity, options), {})
		for table, pks in to_commit.items():
			print('[INFO] Committing on Table ' + table)
			with self.table_service_client.get_table_client(table_name=table) as table_client:
				for _, operations in pks.items():
					# Apparently Azure tables can handle a maximum of 100 lines per batch update
					for op_chunk in self.__chunks(operations, 99):
						try:
							yield table_client.submit_transaction(op_chunk)
						# Apparently some large operation chunks may fail because of transactionerrors
						# the solution is to subdivide the batch and to retry
						except Exception as ex:
							logging.warning(ex)
							for operation in op_chunk:
								try:
									if first(operation, None) == 'upsert':
										# ooooh shit, it's him again
										yield self._unfuck_upsert_error(table_client, *operation)
									else:
										yield table_client.submit_transaction([operation])
								except Exception as ex:
									logging.error(ex)
	
	def _unfuck_upsert_error(self, table_client, operation, entity={}, options={}):
		'''helper method that, for lack of better words, unfucks
		 the fuckery an Upsert operation may cause in Azure Tables'''
		# UPSERT is never transactional, so here we have to try update 
		# and if it fails, we try insert
		try:
			return table_client.update_entity(entity=entity)
		except:
			return table_client.create_entity(entity=entity)

	def _separate_pks(self, pk, in_tuple, dictionary):
		'''helper method to pidgeonhole entities by Partition Key value'''
		if dictionary.get(pk):
			dictionary[pk].append(in_tuple)
		else:
			dictionary[pk] = [in_tuple]
		return dictionary

	def __chunks(self, lst, n):
		"""Yield successive n-sized chunks from lst."""
		for i in range(0, len(lst), n):
			yield lst[i:i + n]

	def is_unique(self, table, value, field='RowKey'):
		'''Tests if a certain value is unique within registered entities

		Parameters
		----------
		table: string
			the table you want to check upon
		
		value: object
			the value you want to checo its uniqueness
		
		field: string, default 'RowKey'
			the table field where the value belongs to, defaults
			to 'RowKey'
		
		Returns
		-------

		bool
			wheter or not the value is unique
		'''
		with self.table_service_client.get_table_client(table) as tc:
			if len(list(tc.query_entities(
						query_filter= str(field)+ ' eq @value', parameters={'value':value}
					))) >0:
				return False
			else:
				return True
	
	def generate_unique_value(self, table, field = 'RowKey', **kwargs):
		'''suggests a unique (not registered yet) value for a given field for a given entity
		
		Parameters
		----------
		table: string
			the table you want to generate a value for
		
		field: string, default 'RowKey'
			the table field where the value would belongs to, defaults
			to 'RowKey'
		
		Returns
		-------

		string
			a unique string value for the considered field
		'''
		temp_id = id_generator(**kwargs)
		while not self.is_unique(table, temp_id, field):
			temp_id = id_generator(**kwargs)
		return temp_id






from io import BytesIO
import numpy as np
import pandas as pd
import os
import shapefile

import shapely as sp
import ssl
import urllib.request
import zipfile

from shutil import rmtree
from sklearn.neighbors import NearestNeighbors, BallTree
from pyproj import Proj, Transformer, transform

from math import *

# important helper function
pRad = 6356.7523142
EqRad = 6378.137

def get_earth_radius(siteLatitude)->float:
    '''
    Returns the Earth's radius at a certain latitude

    Parameters
    ----------
    siteLatitude: float
        the latitude (in degrees) of your location
    '''
    return pRad + (90 - abs(siteLatitude)) / 90 * (EqRad - pRad)

def distance_to_bounds(lon, lat, boundaries):
    '''
    This function will create a numpy array of distances
    to a certain boundary, like a shore. It will contain and ID for AIS points and
    the distance to the nearest coastline point.
    '''
    coastline_coords = boundaries
    tree = BallTree(np.radians(coastline_coords), metric='haversine')
    coords = pd.concat([np.radians(lon), np.radians(lat)], axis=1)
    dist, ind = tree.query(coords, k=1)
    # dists is in rad; convert to km
    df_distance_to_shore = pd.Series(dist.flatten()*get_earth_radius(lat), name='distance_to_bound')
    df_lat = pd.Series(coastline_coords[ind].flatten()[1], name='nearest_bound_point_lat')
    df_lon = pd.Series(coastline_coords[ind].flatten()[0], name='nearest_bound_point_lon')
    return pd.concat([df_distance_to_shore, df_lat, df_lon], axis=1)


def convert_goem(geom, input_projection, output_projection = 'epsg:4326', names=None, shp =None, shx = None, dbf = None, flip_coords=False):
    '''Converts geometries from one projection to another with acceptable error. 
    In the process fixes also a few geometry errors.
    
    Parameters
    ----------
    geom: shapefile.Reader
        The shepefile you need to sanitize.
    
    input_projection: string
        an epsg identifier for the input datum, like 'epsg:4326'
    
    output_projection: string, default
        an epsg identifier for the output datum, like 'epsg:4326'.
        It defaults to 'epsg:4326' because it is WGS-84, aka the GPS
        cooridnates, so it's you are most likely to desire as output.
    
    names: list, default None
        labels you'd like the output gemetries to have in the output
        shapefile object. If None, the output shapefile will have items
        identified by incremental numbers.
    
    shp: File-like object, default None
        The file object you expect your shapefile's shp object
        to be written to. If None, a new BytesIO will be created.
    
    shx: File-like object, default None 
        The file object you expect your shapefile's shx object
        to be written to. If None, a new BytesIO will be created.
    
    dbf: File-like object, default None
        The file object you expect your shapefile's dbf object
        to be written to. If None, a new BytesIO will be created.
    
    flip_coords: bool default False
        if set to True, the function will return mirrored geometries.
        Why? Because -sometimes- shapefile geometries are flipped
        for unfathomable reasons and we may need to fix that.

    Returns
    -------
    shp , shx , dbf
        three file-like objects that collectively make for a new
        Shapefile. You may serialize them or open them with pyshp

    '''
    if shp is None:
        print('Creating shp BytesIO object')
        shp = BytesIO()
    if shx is None:
        print('Creating shp BytesIO object')
        shx = BytesIO()
    if dbf is None:
        print('Creating shp BytesIO object')
        dbf = BytesIO()
    wgs_shp = shapefile.Writer(shp=shp, shx=shx, dbf=dbf)
    wgs_shp.field('name', 'C')
    transformer = Transformer.from_crs(input_projection, output_projection)
    for idx, feature in enumerate(geom):
        # if there is only one part
        if len(feature.parts) == 1:
            print('Converting Polygon')
            # create empty list to store all the coordinates
            poly_list = []
            # get each coord that makes up the polygon
            for coords in feature.points:
                x, y = coords[0], coords[1]
                # tranform the coord
                new_x, new_y = transformer.transform(x, y)
                # put the coord into a list structure
                if flip_coords:
                    poly_coord = [float(new_y), float(new_x)]
                else:
                    poly_coord = [float(new_x), float(new_y)]
                # append the coords to the polygon list
                poly_list.append(poly_coord)
            # add the geometry to the shapefile.
            wgs_shp.poly([poly_list])
        else:
            print('Converting MultiPolygon of ' + str(len(feature.parts)) + ' parts')
            # append the total amount of points to the end of the parts list
            #feature.parts.append(len(feature.points))
            # enpty list to store all the parts that make up the complete feature
            poly_list = []
            # keep track of the part being added
            parts_counter = 0

            # while the parts_counter is less than the amount of parts
            while parts_counter < len(feature.parts) - 1:
                # keep track of the amount of points added to the feature
                coord_count = feature.parts[parts_counter]
                # number of points in each part
                no_of_points = abs(feature.parts[parts_counter] - feature.parts[parts_counter + 1])
                # create list to hold individual parts - these get added to poly_list[]
                part_list = []
                # cut off point for each part
                end_point = coord_count + no_of_points

                # loop through each part
                while coord_count < end_point:
                    for coords in feature.points[coord_count:end_point]:
                        x, y = coords[0], coords[1]
                        # tranform the coord
                        new_x, new_y = transformer.transform(x, y)
                        # put the coord into a list structure
                        if flip_coords:
                            poly_coord = [float(new_y), float(new_x)]
                        else:
                            poly_coord = [float(new_x), float(new_y)]
                        # append the coords to the part list
                        part_list.append(poly_coord)
                        coord_count = coord_count + 1
                # append the part to the poly_list
                poly_list.append(part_list)
                parts_counter = parts_counter + 1
            # add the geometry to to new file
            wgs_shp.poly(poly_list)
        if names is None:
            wgs_shp.record(str(idx))
        else:
            wgs_shp.record([names[idx]])
    wgs_shp.close()
    return shp , shx , dbf

def read_shape_from_zip(package:zipfile.ZipFile, name:str):
	'''SHP files are often packaged as ZIP files, so this function extracts
	Shapes from Zipfiles.
    
    Parameters
    ----------
    package: zipfile.ZipFile
        the zip file, already openend with the zipfile package
    
    name:   string
        the relative base path of your shapefile components, like 'bounds/nuts0'

    Returns
    -------
    shapefile.Reader
        a shapefile object that can be handled with the pyshp package
    '''
	myshp = package.open(name + '.shp')
	mydbf = package.open(name + '.dbf')
	return shapefile.Reader(shp=myshp, dbf=mydbf)

